@@ -59,43 +59,74 @@ class Json:
 
 class PgDatabase:
 
-    _settings = None
+    _settings = dict()
     _connection = None
+    _auto_disconnect = True
     _result = None
 
-    def __init__(self, settings=None):
-        self._settings = settings
+    def __init__(self, **kwargs):
+        self._settings = kwargs.get('settings')
+        self._auto_disconnect = kwargs.get('auto_disconnect', self._auto_disconnect)
+        self._connection = kwargs.get('connection')
+        if self._connection is not None:
+            self._auto_disconnect = False
+        if kwargs.get('schema') is not None:
+            if not isinstance(self._settings, dict):
+                self._settings['schema'] = kwargs.get('schema')
 
     def get_attr(self, attr_name):
+        if not isinstance(self._settings, dict):
+            return None
         return self._settings.get(attr_name)
 
-    def connect(self, dbname=None, close_connection=False):
-        err_str = f'Error in {self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
-        if self._settings is not None:
-            conn_dict = copy.deepcopy(self._settings)
-            if dbname is not None:
-                conn_dict['dbname'] = dbname
-            conn_string = "host='{host}' port={port} dbname='{dbname}' user='{user}' password='{password}'".format(
-                **conn_dict)
-            try:
-                self._connection = psycopg2.connect(conn_string, cursor_factory=psycopg2.extras.RealDictCursor)
-                dec2_float = psycopg2.extensions.new_type(
-                    psycopg2.extensions.DECIMAL.values,
-                    'DEC2FLOAT',
-                    lambda value, curs: float(value) if value is not None else None)
-                psycopg2.extensions.register_type(dec2_float, self._connection)
-                self._connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
-                if close_connection:
-                    self.disconnect()
-                return True
-            except Exception as e:
-                print(f'{err_str}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}')
-                self._connection = None
-                return False
+    def set_attr(self, attr_name, attr_value):
+        if isinstance(self._settings, dict):
+            self._settings[attr_name] = attr_value
 
-    def disconnect(self):
+    @property
+    def connection(self):
+        return self._connection
+
+    def connect(self, dbname=None, close_connection=False, reconnect=False):
         err_str = f'Error in {self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
+        if self._connection is not None and reconnect is False:
+            return True
+        if not isinstance(self._settings, dict):
+            print(f'{err_str}: not isinstance(self._settings, dict)...')
+            return False
+        conn_dict = copy.deepcopy(self._settings)
+        if dbname is not None:
+            conn_dict['dbname'] = dbname
+        conn_string = "host='{host}' port={port} dbname='{dbname}' user='{user}' password='{password}'".format(
+            **conn_dict)
+        try:
+            self._connection = psycopg2.connect(conn_string, cursor_factory=psycopg2.extras.RealDictCursor)
+            dec2_float = psycopg2.extensions.new_type(
+                psycopg2.extensions.DECIMAL.values,
+                'DEC2FLOAT',
+                lambda value, curs: float(value) if value is not None else None)
+            psycopg2.extensions.register_type(dec2_float, self._connection)
+            self._connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
+            if close_connection:
+                self.disconnect()
+            return True
+        except Exception as e:
+            print(f'{err_str}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}')
+            self._connection = None
+            return False
+
+    def connect_to_db(self):
+        self._result = None
         if self._connection is not None:
+            return True
+        self.connect()
+        return False if self._connection is None else True
+
+    def disconnect(self, force=False):
+        err_str = f'Error in {self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
+        if self._connection is None:
+            return
+        if self._auto_disconnect is True or force is True:
             try:
                 self._connection.rollback()
                 self._connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
@@ -116,35 +147,26 @@ class PgDatabase:
 
     def _execute(self, sql_string, sql_params=None, get_result=True):
         err_str = f'Error in {self.__class__.__name__}.{sys._getframe().f_code.co_name}()'
-        self._result = None
-        if self._connection is not None:
-            try:
-                cursor = self._connection.cursor()
-                if sql_params is None:
-                    cursor.execute(sql_string)
-                else:
-                    cursor.execute(sql_string, self._update_dict(sql_params))
-                if get_result:
-                    self._result = cursor.fetchall() or []
-                return True
-            except Exception as e:
-                print(f'{err_str}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}')
-                self._connection.rollback()
-                self._result = None
-                raise e
-                return False
-            finally:
-                if cursor is not None:
-                    cursor.close()
-        else:
+        if not self.connect_to_db():
             return False
-
-    def connect_to_db(self):
-        self._result = None
-        if self._connection is not None:
+        try:
+            cursor = self._connection.cursor()
+            if sql_params is None:
+                cursor.execute(sql_string)
+            else:
+                cursor.execute(sql_string, self._update_dict(sql_params))
+            if get_result:
+                self._result = cursor.fetchall() or []
             return True
-        self.connect()
-        return False if self._connection is None else True
+        except Exception as e:
+            print(f'{err_str}: {type(e)}: {str(e)}; traceback: {traceback.print_exc()}')
+            self._connection.rollback()
+            self._result = None
+            raise e
+            return False
+        finally:
+            if cursor is not None:
+                cursor.close()
 
     def is_in_transaction(self):
         if self._connection is None:
@@ -177,19 +199,10 @@ class PgDatabase:
             self._connection.rollback()
             self._connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
 
-    def get_connection(self):
-        if not self.connect_to_db():
-            return None
-        return self._connection
-
     def _execute_sql(self, sql_string, sql_params=None, get_result=True, returning=None):
         self._result = None
-        close_connection = True if self._connection is None else False
-        if not self.connect_to_db():
-            return None
         res = self._execute(sql_string, sql_params=sql_params, get_result=get_result)
-        if close_connection:
-            self.disconnect()
+        self.disconnect()
         if returning:
             if not isinstance(self._result, list):
                 return None
@@ -360,7 +373,6 @@ class PgDatabase:
                 self._connection.rollback()
                 return False
 
-        close_connection = True if self._connection is None else False
         if not self.connect_to_db():
             return False
         cursor = self._connection.cursor()
@@ -405,9 +417,8 @@ class PgDatabase:
             if os.path.exists(csv_filepath):
                 os.remove(csv_filepath)
 
-        self._connection.commit()
-        if close_connection:
-            self.disconnect()
+        self.commit()
+        self.disconnect()
 
     def insert_df(self, table, df):
         df_columns = list(df.columns.values)
@@ -419,9 +430,8 @@ class PgDatabase:
             return False
         cursor = self._connection.cursor()
         psycopg2.extras.execute_batch(cursor, insert_stmt, df.values)
-        self._connection.commit()
-        if close_connection:
-            self.disconnect()
+        self.commit()
+        self.disconnect()
 
     def create_database(self):
         sql_string = 'CREATE DATABASE {dbname};'.format(**self._settings)
@@ -623,15 +633,20 @@ class TasksDatabase:
         return True
 
     @classmethod
-    def init(cls, database_dict):
+    def init(cls, **kwargs):
         err_str = f'Error in {cls.__name__}.{sys._getframe().f_code.co_name}()'
-        pg_database = PgDatabase(database_dict)
+        pg_database = PgDatabase(**kwargs)
         if pg_database.connect(close_connection=True):
             TasksDatabase.pg_database = pg_database
             return True
-        print(f'{err_str}: database "{database_dict.get("dbname")}" not available')
+        print(f'{err_str}: database not available')
         return False
 
+    @classmethod
+    def force_disconnect(cls):
+        if TasksDatabase.pg_database is not None:
+            TasksDatabase.pg_database.disconnect(force=True)
+
     @classmethod
     def add_task(cls, task_dict):
         func_name = f'{cls.__name__}.{sys._getframe().f_code.co_name}()'
